/************************************************
 * This code is free to use, but this copyright
 * should be deleted, otherwise, they won't 
 * think you wrote it!
 * 
 * If you find this useful, please email me at
 * aprenot@aaronprenot.com
 * 
 * (c)2004 - Aaron Prenot
 * http://www.aaronprenot.com
 * **********************************************/
using System;
using System.Collections.Generic;

namespace Diff
{
	/// <summary>
	/// 
	/// </summary>
	public class DiffBitMatrix : Diff.BitMatrix
	{
		protected ComparableStreamReader	_source;
		protected ComparableStreamReader	_destination;
		protected IComparable[]	_bufSource;
		protected IComparable[]	_bufDestination;
		protected int			_windowStartX, _windowStartY;	// the offset within the stream that the window starts at
		protected int			_windowMaxX, _windowMaxY;		// the max x and y location within the current window
		protected int			_maxX, _maxY;					// the max x and y locations for each stream.
		protected bool			_foundEndX, _foundEndY = false;

		public new bool this[int x, int y]
		{
			get 
			{ 
				if ( x >= _windowMaxX || y >= _windowMaxY ) Slide(x,y);
				x -= _windowStartX;
				y -= _windowStartY;
				return GetBit(x, y);
			}
			set 
			{ 
				if ( x >= _windowMaxX || y >= _windowMaxY ) Slide(x,y);
				x -= _windowStartX;
				y -= _windowStartY;
				SetBit(x, y, value); 
			}
		}

		public DiffBitMatrix(int square) : this(square, square)
		{
		}

		public DiffBitMatrix(int width, int height) : base(width, height)
		{
			this._bufSource = new IComparable[width];
			this._bufDestination = new IComparable[height];
			this._windowStartX = this._windowStartY = 0;
			this._windowMaxX = this._windowMaxY = 0;
			this._maxX = this._maxY = int.MaxValue;
		}

		public DiffBitMatrix(KeyValuePair<int,int> size) : this(size.Key, size.Value)
		{
		}

		public int MaxX
		{
			get { return _maxX; }
		}

		public int MaxY
		{
			get { return _maxY; }
		}

		public int WindowMaxX
		{
			get { return _windowMaxX; }
		}

		public int WindowMaxY
		{
			get { return _windowMaxY; }
		}

		public ComparableStreamReader Source
		{
			get { return _source; }
			set { _source = value; }
		}

		public ComparableStreamReader Destination
		{
			get { return _destination; }
			set { _destination = value; }
		}

		public IComparable SourceValue(int X)
		{
			if ( X < _windowStartX || X > _windowMaxX ) throw new ArgumentOutOfRangeException("X");
			return _bufSource[(X - _windowStartX)];
		}

		public IComparable DestinationValue(int Y)
		{
			if ( Y < _windowStartY || Y > _windowMaxY ) throw new ArgumentOutOfRangeException("Y");			
			return _bufDestination[(Y - _windowStartY)];
		}

		public void Start()
		{
			Slide(0,0);
		}

		public void Slide(int X, int Y)
		{
			if ( _foundEndX && _foundEndY ) return;
			int i;
			IComparable[] bufSource = new IComparable[_size.Key];
			IComparable[] bufDestination = new IComparable[_size.Value];
			// first, read in the remaining values from the previous buffer
			for ( i = X; i < _windowMaxX; i++) bufSource[(i - X)] = _bufSource[(i - _windowStartX)];
			for ( i = Y; i < _windowMaxY; i++) bufDestination[(i - Y)] = _bufDestination[(i - _windowStartY)];
			// reset the matrix start points
			_windowStartX = X;
			_windowStartY = Y;

			// now read in the x lines
			if ( !_foundEndX )
			{
				for ( i = (_windowMaxX - X); i < _size.Key; i++) 
				{
					if ( (bufSource[i] = _source.GetNext()) == null )
					{
						_windowMaxX = i + _windowStartX;
						_maxX = i + _windowStartX;
						_foundEndX = true;
						break;
					}
				}
			}
			if ( !_foundEndX ) _windowMaxX = _windowStartX + _size.Key;

			// now read in the y lines
			if ( !_foundEndY )
			{
				for ( i = (_windowMaxY - Y); i < _size.Value; i++)
				{
					if ( (bufDestination[i] = _destination.GetNext()) == null )
					{
						_windowMaxY = i + _windowStartY;
						_maxY = i + _windowStartY;
						_foundEndY = true;
						break;
					}
				}
			}
			if ( !_foundEndY ) _windowMaxY = _windowStartY + _size.Value;
			
			Clear();
			// now do the comparison
			for ( i = 0; i < (_windowMaxX - _windowStartX); i++ )
			{
				for ( int j = 0; j < (_windowMaxY - _windowStartY); j++ )
				{
					int result = bufSource[i].CompareTo(bufDestination[j]);
					if ( result == 0 ) SetBit(i,j, true);
					else SetBit(i,j, false);
				}
			}
			_bufSource = bufSource;
			_bufDestination = bufDestination;
		}

		/// <summary>
		/// Returns the longest diagonal on the x or y line if 2 diagonals of the same length
		/// exist, it returns the one that starts closest to the origin, x first.
		/// </summary>
		/// <param name="pt">The X and Y Origins on which it should search.</param>
		/// <returns></returns>
		public KeyValuePair<int,int> GetDiagonal(KeyValuePair<int,int> pt)
		{
			return GetDiagonal(pt.Key, pt.Value);
		}
		
		/// <summary>
		/// Returns the longest diagonal on the x or y line if 2 diagonals of the same length
		/// exist, it returns the one that starts closest to the origin, x first.
		/// </summary>
		/// <param name="x">The X Origin on which it should search.</param>
		/// <param name="y">The Y Origin on which it should search.</param>
		/// <returns></returns>
		public KeyValuePair<int,int> GetDiagonal(int x, int y)
		{
			KeyValuePair<int,int> maxDiag = new KeyValuePair<int,int>(-1,-1);
			int	maxDiagLength = 0;
			int i;

			int dx, dy;
			int count;

			// look down the x side
			if ( !( x < _windowMaxX) ) Slide(x,y);
			for ( i = x; i < _windowMaxX; i++)
			{
				if ( this[i, y] )
				{
					dx = i;
					dy = y;
					// found a match.  now follow it for its length
					count = 0;
					while ( this[dx++, dy++] )
					{
						count++;
						if ( !(dx < _windowMaxX && dy < _windowMaxY) ) break;
					}
					if ( count > maxDiagLength && count > 0 /* who cares about 1 matched character? */ ) 
					{
						maxDiag = new KeyValuePair<int, int> (i, y);
						maxDiagLength = count;
					}
				}
			}
			
			// look down the y side
			if ( !(y < _windowMaxY) ) Slide(x,y);
			for ( i = y; i < _windowMaxY; i++)
			{
				if ( this[x, i] )
				{
					dx = x;
					dy = i;
					// found a match.  now follow it for its length
					count = 0;
					while ( this[dx++, dy++] )
					{
						count++;
						if ( !(dx < _windowMaxX && dy < _windowMaxY) ) break;
					}
					if ( (count > maxDiagLength && count > 0) || ( (count == maxDiagLength && count > 0) && ((maxDiag.Key > -1) ? ((i - y) < (maxDiag.Value - x)) : true)) )
					{
						maxDiag = new KeyValuePair<int, int> (x, i);
						maxDiagLength = count;
					}
				}
			}
			return maxDiag;
		}
	}
}
