/************************************************
 * This code is free to use, but this copyright
 * should be deleted, otherwise, they won't 
 * think you wrote it!
 * 
 * If you find this useful, please email me at
 * aprenot@aaronprenot.com
 * 
 * (c)2004 - Aaron Prenot
 * http://www.aaronprenot.com
 * **********************************************/

using System;
using System.Collections.Generic;

namespace Diff
{
	/// <summary>
	/// The work horse of the algorithm.
	/// </summary>
	public class DiffEngine
	{
		protected DiffBitMatrix		_bm;			// the matrix of comparisons;
		protected int			_windowSize;	// the buffer size
		protected int			_x, _y;			// the location within the streams;
		protected int			_origX, _origY;	// the location the current operation started on
		protected ComparableStreamReader	_source;
		protected ComparableStreamReader	_destination;
		
		protected EditScript	_editScript;
		protected bool			_inDiag;

		public DiffEngine()
		{
			this._bm = null;
			this._windowSize = 256;
			this._x = this._y = 0;
			this._origX = this._origY = 0;
			this._source = null;
			this._destination = null;

			this._editScript = new EditScript();
			this._inDiag = false;
		}

		public int WindowSize
		{
			get { return _windowSize; }
			set { _windowSize = value; }
		}

		public EditScript Script
		{
			get { return _editScript; }
		}

		public EditScript Compare(ComparableStreamReader source, ComparableStreamReader destination)
		{
			_bm = new DiffBitMatrix(_windowSize);
			_bm.Source = source;
			_bm.Destination = destination;
			_bm.Start();
			int count;
			DiffCommand dc;
			while ( _x < _bm.MaxX && _y < _bm.MaxY )
			{
				if ( _inDiag )
				{
					if ( !_bm[_x, _y] )
					{
						// insert skip command
						dc = new DiffCommand(DiffCommandType.Skip, null, (_x - _origX));
						_editScript.Add(dc);

						// change state
						_origX = _x; _origY = _y;
						_bm.Slide(_x, _y);	// we are done up to this point, so lets get some fresh data
						_inDiag = false;
						continue;
					}
					else
					{
						_x++; _y++;
					}
				}
				else
				{
					KeyValuePair<int,int> diag = _bm.GetDiagonal(_x, _y);
					if ( diag.Key != -1 )	// found a diagonal
					{
						if ( diag.Value != _y ) // the diag is on the x axis, so we must delete to get to it
						{
							for ( ; _y < diag.Value; _y++)
							{
								dc = new DiffCommand(DiffCommandType.Insert, _bm.DestinationValue(_y), -1);
								_editScript.Add(dc);
							}
						}
						if ( diag.Key != _x ) // the diag is on the y axis, so we must insert to get to it.
						{
							count = 0;
							for ( ; _x < diag.Key; _x++) count++;
							dc = new DiffCommand(DiffCommandType.Delete, null, count);
							_editScript.Add(dc);
						}
						// change state
						_origX = _x; _origY = _y;
						_bm.Slide(_x,_y);	// we are done up to this point, so lets get some fresh data
						_inDiag = true;
						continue;
					}
					else	// didn't find a diagonal, so process as normal
					{
						if ( _x < _bm.MaxX && _y < _bm.MaxY )	// this is for some special cases where we don't realize that we are at the end until this point
						{
							dc = new DiffCommand(DiffCommandType.Delete, null, 1);
							_editScript.Add(dc);
							dc = new DiffCommand(DiffCommandType.Insert, _bm.DestinationValue(_y), -1);
							_editScript.Add(dc);
							_x++; _y++;
						}
					}
				}
			}
			// save state
			if ( _inDiag )
			{
				dc = new DiffCommand(DiffCommandType.Skip, null, (_x - _origX));
				_editScript.Add(dc);
			}

			if ( _x < _bm.MaxX )
			{
				// mass deletes
				count = 0;
				for ( ; _x < _bm.WindowMaxX; _x++ ) count++;	// eat up the remaining bytes in the window
				while ( source.GetNext() != null ) count++;		// eat up the remaining bytes in the stream
				dc = new DiffCommand(DiffCommandType.Delete, null, count);
				_editScript.Add(dc);
			}
			if ( _y < _bm.MaxY )
			{
				// mass inserts
				for ( ; _y < _bm.WindowMaxY; _y++)	// eat up the remaining bytes in the window
				{
					dc = new DiffCommand(DiffCommandType.Insert, _bm.DestinationValue(_y), -1);
					_editScript.Add(dc);
				}
				while ( source.GetNext() != null )	// eat up the remaining bytes in the stream
				{
					dc = new DiffCommand(DiffCommandType.Insert, _bm.DestinationValue(_y), -1);
					_editScript.Add(dc);
				}
			}
			return _editScript;
		}
	}
}
